package ac.ed.lurg.country;

import java.util.HashMap;
import java.util.Map;

import ac.ed.lurg.Timestep;
import ac.ed.lurg.country.gams.GamsCountryInput;
import ac.ed.lurg.country.gams.GamsRasterInput;
import ac.ed.lurg.country.gams.GamsRasterOptimiser;
import ac.ed.lurg.country.gams.GamsRasterOutput;
import ac.ed.lurg.demand.DemandManager;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.landuse.IrrigationItem;
import ac.ed.lurg.landuse.LandCoverItem;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.YieldRaster;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class CountryAgent {
	
	private DemandManager demandManager;
	private CompositeCountry country;
	
	private Map<Timestep, GamsRasterOutput> resultsTimeseries = new HashMap<Timestep, GamsRasterOutput>();
	private Timestep currentTimestep;
	private YieldRaster countryYieldSurfaces;
	private Map<CommodityType, Double> currentProjectedDemand;
	private Map<CropType, Double> tradeBarriers;
	
	public CountryAgent(DemandManager demandManager,CompositeCountry country, RasterSet<LandCoverItem> initialLC,
			Map<CropType, CropUsageData> cropUsageData,  Map<CropType, Double> tradeBarriers) {
		
		this.demandManager = demandManager;
		this.country = country;
		RasterSet<LandUseItem> cropAreaRaster = convertInitialLC(initialLC);
		this.tradeBarriers = tradeBarriers;
		
		GamsRasterOutput initialData = new GamsRasterOutput(cropAreaRaster, cropUsageData);
		resultsTimeseries.put(new Timestep(0), initialData);
	}

	private RasterSet<LandUseItem> convertInitialLC(RasterSet<LandCoverItem> initialLC) {
		RasterSet<LandUseItem> landUseRaster = new RasterSet<LandUseItem>(initialLC.getHeaderDetails());
		
		for (Map.Entry<RasterKey, LandCoverItem> entry : initialLC.entrySet()) {
			LandUseItem areasItem = new LandUseItem();
			RasterKey key = entry.getKey();
			areasItem.setLandCoverAreas(entry.getValue());
			areasItem.setCropFraction(CropType.WHEAT, 0.5); // random start, better if we could get data, but free substitution between crops so not critical
			areasItem.setCropFraction(CropType.MAIZE, 0.5);
			areasItem.setProtectedArea(entry.getValue());
			landUseRaster.put(key, areasItem);
		}
		
		return landUseRaster;
	}
	
	public CompositeCountry getCountry() {
		return country;
	}

	public GamsRasterOutput determineProduction(Timestep timestep, YieldRaster countryYieldSurfaces, RasterSet<IrrigationItem> irrigData, Map<CropType, GlobalPrice> worldPrices) {
		currentTimestep = timestep;
		this.countryYieldSurfaces = countryYieldSurfaces;
		
		// get projected demand
		currentProjectedDemand = demandManager.getDemand(country, timestep.getYear());
		
		if (currentProjectedDemand.size() == 0) {
			LogWriter.printlnError("No demand for country " + country + " so skipping it");
		}
		else if (countryYieldSurfaces.size() == 0 ) {
			LogWriter.printlnError("No yield values for " + country + " so skipping it");
		}
		else {
			// optimize areas and intensity 
			GamsRasterInput input = getGamsRasterInput(currentProjectedDemand, tradeBarriers, worldPrices, irrigData);
			GamsRasterOptimiser opti = new GamsRasterOptimiser(input);
			LogWriter.println("Running " + country.getName() + ", currentTimestep " + currentTimestep);
			
			GamsRasterOutput result = opti.run();
			resultsTimeseries.put(timestep, result);
			return result;
		}
		
		return null;
	}
	
	public Map<CommodityType, Double> getCurrentProjectedDemand() {
		return currentProjectedDemand;
	}

	private GamsRasterInput getGamsRasterInput(Map<CommodityType, Double> projectedDemand, Map<CropType, Double> tradeBarriers, Map<CropType, GlobalPrice> worldPrices, RasterSet<IrrigationItem> irrigData) {
//TODO why pass in projectedDemand when currentProjectedDemand belongs to CountryAgent class already?
		GamsRasterOutput prevOutput;
		Map<CropType, Double> cropAdjs;
		boolean calibrate;

		if (currentTimestep.isInitialTimestep()) {  // initialisation time-step
			calibrate = true;
			prevOutput = resultsTimeseries.get(currentTimestep);
			cropAdjs = null;
		}
		else { // normal (not the initial) time-step
			calibrate = false;
			prevOutput = resultsTimeseries.get(currentTimestep.getPreviousTimestep());
			cropAdjs = prevOutput.getCropAdjustments();
		}
		
		Map<CropType, Double> baseNetImport = new HashMap<CropType, Double>();
		Map<CropType, Double> maxOfProdOrSupply = new HashMap<CropType, Double>();

		for (Map.Entry<CropType, CropUsageData> entry : prevOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			double netImports = cropUsage.getNetImports();
			baseNetImport.put(entry.getKey(), netImports);
			maxOfProdOrSupply.put(entry.getKey(), cropUsage.getProduction() + Math.max(netImports, 0));
		}
				
		GamsCountryInput countryLevelInputs = GamsCountryInput.createInput(country, projectedDemand, tradeBarriers, worldPrices, baseNetImport, maxOfProdOrSupply, cropAdjs, calibrate);	
		GamsRasterInput input = new GamsRasterInput(currentTimestep, countryYieldSurfaces, prevOutput.getLandUses(), irrigData, countryLevelInputs);

		return input;
	}
}