package ac.ed.lurg.landuse;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.types.CropToDouble;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.Interpolator;
import ac.sac.raster.InterpolatingRasterItem;

public class LandUseItem implements InterpolatingRasterItem<LandUseItem> {

	private Map<CropType, Intensity> intensityMap = new HashMap<CropType, Intensity>();
	private Map<CropType, Double> cropFractions = new HashMap<CropType, Double>();
	private Map<LandCoverType, Double> landCoverAreas = new HashMap<LandCoverType, Double>();
	private double protectedArea; //protected area in Mha
	
	public void setProtectedArea(LandCoverItem landCover){
		if (landCover != null) {
			double areaTotal = this.getTotalLandCoverArea();
			this.protectedArea = landCover.getProtectedArea()*areaTotal;
		}
	}
	
	public double getProtectedArea(){
		return this.protectedArea;
	}
	
	public Intensity getIntensity(CropType crop) {
		return intensityMap.get(crop);
	}
	
	public void setIntensity(CropType crop, Intensity intensityData) {
		intensityMap.put(crop, intensityData);
	}
	
	/** Fertiliser rate in kg/ha */
	public double getFertiliserRate(CropType crop) {
		Intensity i = getIntensity(crop);
		return (i == null) ? 0 : i.getFertiliserAmount();
	}
	
	/** Fertiliser amount in kt (1000 tonnes), for this location */
	public double getFertiliserAmount(CropType c) {
		double rate = getFertiliserRate(c);
		double area = getCropArea(c);
		return rate * area;
	}
	
	public double getFertiliserAverageRate(CropType... crops) {
		double fertTotal = 0;
		double areaTotal = 0;
		
		for (CropType c : crops) {
			fertTotal += getFertiliserAmount(c);
			areaTotal = getCropArea(c);
		}
		
		return areaTotal > 0 ? fertTotal / areaTotal : 0;
	}

	
	public static double getFertiliserTotal(Collection<? extends LandUseItem> items, CropType... crops) {			
		double total = 0;
		for (LandUseItem a : items) {
			if (a == null)
				continue;
			
			for (CropType c : crops)
				total += a.getFertiliserAmount(c);
		}
			
		return total;
	}

	public static double getFertiliserTotal(Collection<? extends LandUseItem> items, Collection<CropType> crops) {
		return getFertiliserTotal(items, crops.toArray(new CropType[crops.size()]));
	}

	/** Irrigation rate in litre/m2 */
	public double getIrrigationRate(CropType crop) {
		Intensity i = getIntensity(crop);
		
		return (i == null) ? 0 : i.getIrrigationRate();
	}

	/** Irrigation amount in km3 or 10^9 m3, for this location */
	private double getIrrigationAmount(CropType c) {
		double rate = getIrrigationRate(c);
		double area = getCropArea(c);
		return rate * area * 0.01;  // rate(10^-3m or mm or l/m2) * area(10^6ha) = 10^3 m.ha = 10^7 m3 = 1/100 km3
	}

	/** Irrigation rate in litre/m2 */
	public double getIrrigationAverageRate(CropType... crops) {
		double irrigTotal = 0;
		double areaTotal = 0;
		
		for (CropType c : crops) {
			double area = getCropArea(c);
			irrigTotal += getIrrigationRate(c) * area;
			areaTotal += area;
		}
		
		return areaTotal > 0 ? irrigTotal / areaTotal : 0;
	}
	
	/** Irrigation amount in km3, for this location */
	public static double getIrrigationTotal(Collection<? extends LandUseItem> items, CropType... crops) {			
		double total = 0;
		for (LandUseItem a : items) {
			if (a == null)
				continue;
			
			for (CropType c : crops)
				total += a.getIrrigationAmount(c);
		}
			
		return total;
	}

	public static double getIrrigationTotal(Collection<? extends LandUseItem> items, Collection<CropType> crops) {
		return getIrrigationTotal(items, crops.toArray(new CropType[crops.size()]));
	}
	
	

	public void setLandCoverAreas(LandCoverItem landCover) {
		if (landCover != null) {
			for (LandCoverType lcType : LandCoverType.values())
				landCoverAreas.put(lcType, landCover.getLandCoverArea(lcType));
		}
	}
	
	public void setLandCoverAreas(LandUseItem otherAreasItem) {
		for (Entry<LandCoverType, Double> entry : otherAreasItem.landCoverAreas.entrySet()) {
			landCoverAreas.put(entry.getKey(), entry.getValue());
		}
	}

	/** move areas from one land cover to another, return an residual not possible */
	public double moveAreas(LandCoverType toType, LandCoverType fromType, double changeReq) {
		
		double prevTo = getLandCoverArea(toType);
		double prevFrom = getLandCoverArea(fromType);		
		double actuallyChanged = Math.min(prevFrom, changeReq);
		
		setLandCoverArea(toType, prevTo + actuallyChanged);
		setLandCoverArea(fromType, prevFrom - actuallyChanged);

		return changeReq - actuallyChanged;
	}


	public double getCropArea(CropType c) {
		Double d = getLandCoverArea(LandCoverType.CROPLAND) * getCropFraction(c);
		return d;
	}
	
	public double getCropFraction(CropType c) {
		Double d = cropFractions.get(c);
		return d == null ? 0.0 : d;
	}

	public double getCropFraction(CropType... cropsToFind) {
		double totalFract = 0;
		for (CropType crop : cropsToFind) {
			totalFract += getCropFraction(crop);
		}
		return totalFract;
	}

	
	public void setCropFraction(CropType c, double area) {
		cropFractions.put(c, area);
	}
	
	public double getLandCoverArea(LandCoverType c) {
		Double d = landCoverAreas.get(c);
		return d == null ? 0.0 : d;
	}
	
	public void setLandCoverArea(LandCoverType c, double d) {
		if (Double.isNaN(d) || Double.isInfinite(d))
			throw new RuntimeException("AreasItem for " + c + " is " + d);
		
		landCoverAreas.put(c, d);
	}
	
	public double getTotalLandCoverArea() {
		double d = 0;
		for (LandCoverType l : LandCoverType.values()) 
			d += getLandCoverArea(l);
	
		return d;
	}
		
	public double getSuitableLand() {
		double d = 0;

		double protectedTotal = ModelConfig.PROTECTED_AREAS_ENABLED ? protectedArea : 0;
		double totalNatural  = getLandCoverArea(LandCoverType.OTHER_NATURAL) + getLandCoverArea(LandCoverType.FOREST);
		
		if (totalNatural >= protectedTotal)
			d += totalNatural - protectedTotal;

			d += getLandCoverArea(LandCoverType.PASTURE);
			d += getLandCoverArea(LandCoverType.CROPLAND);

		return d;
	}
	
	public double getForestToNaturalFraction() {
		double forest = getLandCoverArea(LandCoverType.FOREST);
		double natural = forest + getLandCoverArea(LandCoverType.OTHER_NATURAL);
		double d = forest / natural;
		return (Double.isNaN(d) || Double.isInfinite(d)) ? 1.0 : d;
	}

	public CropToDouble getCropChanges(LandUseItem prevAreaAggItem) {
		CropToDouble changes = new CropToDouble();
		
		for (CropType c : CropType.getCropsLessPasture()) {
			double change = getCropArea(c) - prevAreaAggItem.getCropArea(c);
			changes.put(c, change);
		}
		
		return changes;
	}

	private boolean isZeroOrNull(Double d) {
		return d == null || d == 0;
	}
	 
	@Override
	public void interpolateAll(LandUseItem fromItem, LandUseItem toItem, double factor) {		
		cropFractions = new HashMap<CropType, Double>();
		landCoverAreas = new HashMap<LandCoverType, Double>();

		Double fromCropCover = fromItem.landCoverAreas.get(LandCoverType.CROPLAND);
		Double toCropCover = toItem.landCoverAreas.get(LandCoverType.CROPLAND);

		if (!isZeroOrNull(fromCropCover) && isZeroOrNull(toCropCover)) { // if start with crop but end with none, take starting crop fractions
			cropFractions.putAll(fromItem.cropFractions);
			intensityMap.putAll(fromItem.intensityMap);
		}
		else if (isZeroOrNull(fromCropCover) && !isZeroOrNull(toCropCover)) { // if start with no crop but end with some, take end crop fractions
			cropFractions.putAll(toItem.cropFractions);
			intensityMap.putAll(toItem.intensityMap);
		}
		else { // otherwise we need to interpolate crop fractions
			for (CropType crop : CropType.values()) {
				Double from = fromItem.cropFractions.get(crop);
				Double to = toItem.cropFractions.get(crop);
				Double d = Interpolator.interpolate(from, to, factor);
				cropFractions.put(crop, d);
				
				Intensity fromIntensity = fromItem.intensityMap.get(crop);
				Intensity toIntensity = toItem.intensityMap.get(crop);
				Intensity interpolateIntensity = toIntensity;  // might still be null
				
				if (fromIntensity != null && toIntensity != null)
					interpolateIntensity = new Intensity(fromIntensity, toIntensity, factor);  // both non-null really interpolate
				else if (fromIntensity != null)
					interpolateIntensity = fromIntensity; // just fromIntensity non-null
				
				intensityMap.put(crop, interpolateIntensity);
			}
		}
		
		for (LandCoverType landCover : LandCoverType.values()) {
			Double from = fromItem.landCoverAreas.get(landCover);
			Double to = toItem.landCoverAreas.get(landCover);
			Double d = Interpolator.interpolate(from, to, factor);
			landCoverAreas.put(landCover, d);
		}
	}
	
	public static double getTotalLandCover(Collection<? extends LandUseItem> landUses, LandCoverType landCover) {
		double total = 0;
		for (LandUseItem a : landUses)
			total += a.getLandCoverArea(landCover);
		
		return total;
	}

	public static double getTotalCropArea(Collection<LandUseItem> landUses, CropType crop) {
		double total = 0;
		for (LandUseItem a : landUses)
			total += a.getCropArea(crop);
		
		return total;
	}
}
