package ac.ed.lurg.output;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map.Entry;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.YieldRaster;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class LpjgOutputer {

	private RasterSet<LandUseItem> landUseRaster;
	private YieldRaster yieldSurfaces;
	private int year;

	public LpjgOutputer(int year, RasterSet<LandUseItem> landUseRaster, YieldRaster yieldSurfaces) {
		this.year = year;
		this.landUseRaster = landUseRaster;
		this.yieldSurfaces = yieldSurfaces;
	}

	public void writeOutput() {
		File outputDir = getOutputDir(year);

		writeLandCoverAndCrop(outputDir);
		writeIntensity(outputDir);
		writeMarkerFile(year, false);
	}

	private static File getOutputDir(int year) {
		String outputDirName = ModelConfig.OUTPUT_DIR + File.separator + year;
		File outputDir = new File(outputDirName);
		outputDir.mkdirs();
		return outputDir;
	}

	private void writeIntensity(File outputDir) {
		BufferedWriter fertWriter = null;
		BufferedWriter irrigWriter = null;

		double refactorMePlease;
		
		try {
			String landCoverFileName = outputDir.getPath() + File.separator + "Fert.txt";
			fertWriter = new BufferedWriter(new FileWriter(landCoverFileName, false));
			fertWriter.write("Lon Lat Year TeWWirr TeSWirr TeCoirr TrRiirr Pasture");
			fertWriter.newLine();

			String cropFractionFileName = outputDir.getPath() + File.separator + "Irrig.txt";
			irrigWriter = new BufferedWriter(new FileWriter(cropFractionFileName, false));
			irrigWriter.write("Lon Lat Year TeWWirr TeSWirr TeCoirr TrRiirr Pasture");
			irrigWriter.newLine();

			for (Entry<RasterKey, LandUseItem> entry : landUseRaster.entrySet()) {
				RasterKey key = entry.getKey();
				LandUseItem item = entry.getValue();
				
				if (item == null)
					continue;
				
				double lat = landUseRaster.getXCoordin(key);
				double lon = landUseRaster.getYCoordin(key);
				boolean isSpringWheat = yieldSurfaces.isSpringWheat(key);

				double winterWheatIrrig = item.getIrrigationAverageRate(CropType.OILCROPS, isSpringWheat ? null : CropType.WHEAT);
				double springWheatIrrig = item.getIrrigationAverageRate(CropType.PULSES, CropType.STARCHY_ROOTS, isSpringWheat ? CropType.WHEAT : null);
				double cornIrrig = item.getIrrigationAverageRate(CropType.MAIZE);
				double riceIrrig = item.getIrrigationAverageRate(CropType.RICE);
				double pastureIrrig = item.getIrrigationAverageRate(CropType.PASTURE);
				irrigWriter.write(String.format("%.2f %.2f %d %.14f %.14f %.14f %.14f %.14f", lat, lon, year, winterWheatIrrig, springWheatIrrig, cornIrrig, riceIrrig, pastureIrrig));
				irrigWriter.newLine();

				double winterWheatFert = item.getFertiliserAverageRate(CropType.OILCROPS, isSpringWheat ? null : CropType.WHEAT);
				double springWheatFert = item.getFertiliserAverageRate(CropType.PULSES, CropType.STARCHY_ROOTS, isSpringWheat ? CropType.WHEAT : null);
				double cornFert = item.getFertiliserAverageRate(CropType.MAIZE);
				double riceFert = item.getFertiliserAverageRate(CropType.RICE);
				double pastureFert = item.getFertiliserAverageRate(CropType.PASTURE);
				fertWriter.write(String.format("%.2f %.2f %d %.14f %.14f %.14f %.14f %.14f", lat, lon, year, winterWheatFert, springWheatFert, cornFert, riceFert, pastureFert));
				fertWriter.newLine();
			}
		}
		catch (IOException e) {
			LogWriter.print(e);
		}
		finally {
			if (fertWriter != null) {
				try {
					fertWriter.close();
				} 
				catch (IOException e) {
					LogWriter.print(e);
				}
			}
			if (irrigWriter != null) {
				try {
					irrigWriter.close();
				} 
				catch (IOException e) {
					LogWriter.print(e);
				}
			}
		}
	}

	public static void writeMarkerFile(int year, boolean errorStatus) {
		File outputDir = getOutputDir(year);

		File markerFile = new File(outputDir.getPath() + File.separator + (errorStatus ? "error" : "done"));
		
		try {
			if (!markerFile.exists())
				new FileOutputStream(markerFile).close();
			markerFile.setLastModified(System.currentTimeMillis());
		}
		catch (IOException e) {
			LogWriter.printlnError("Problem writing marker file");
		}	
	}

	public void writeLandCoverAndCrop(File outputDir) {
		BufferedWriter landCoverWriter = null;
		BufferedWriter cropFractWriter = null;

		try {
			String landCoverFileName = outputDir.getPath() + File.separator + "LandCoverFract.txt";
			landCoverWriter = new BufferedWriter(new FileWriter(landCoverFileName, false));
			landCoverWriter.write("Lon Lat Year CROPLAND PASTURE NATURAL BARREN");
			landCoverWriter.newLine();

			String cropFractionFileName = outputDir.getPath() + File.separator + "CropFract.txt";
			cropFractWriter = new BufferedWriter(new FileWriter(cropFractionFileName, false));
			cropFractWriter.write("Lon Lat Year TeWWirr TeSWirr TeCoirr TrRiirr");
			cropFractWriter.newLine();

			for (Entry<RasterKey, LandUseItem> entry : landUseRaster.entrySet()) {
				RasterKey key = entry.getKey();
				LandUseItem item = entry.getValue();
				double lat = landUseRaster.getXCoordin(key);
				double lon = landUseRaster.getYCoordin(key);

				double expectedArea = landUseRaster.getAreaMha(key);
				double area = item.getTotalLandCoverArea();
				
				if (area > 0 && Math.abs((expectedArea-area)/expectedArea) > 0.01) {  // zero area, due to data not being found, already reported so ignored here
					LogWriter.printlnError("Land cover areas look strange " + key + ": expected=" + expectedArea + ", actual=" + area);
				}
				
				double crop = item.getLandCoverArea(LandCoverType.CROPLAND);
				double pasture = item.getLandCoverArea(LandCoverType.PASTURE);
				double forest = item.getLandCoverArea(LandCoverType.FOREST);
				double otherNatural = item.getLandCoverArea(LandCoverType.OTHER_NATURAL);
				double barren = item.getLandCoverArea(LandCoverType.BARREN);
				landCoverWriter.write(String.format("%.2f %.2f %d %.14f %.14f %.14f %.14f", lat, lon, year, crop/area, pasture/area, (forest+otherNatural)/area, barren/area));
				landCoverWriter.newLine();

				boolean isSpringWheat = yieldSurfaces.isSpringWheat(key);

				double lpjWinterWheatFrac = item.getCropFraction(isSpringWheat ? null : CropType.OILCROPS, isSpringWheat ? null : CropType.WHEAT);
				double lpjSpringWheatFrac = item.getCropFraction(CropType.PULSES, CropType.STARCHY_ROOTS, isSpringWheat ? CropType.WHEAT : null, isSpringWheat ? CropType.OILCROPS : null);
				double lpjCornFrac = item.getCropFraction(CropType.MAIZE);
				double lpjRiceFrac = item.getCropFraction(CropType.RICE);
				cropFractWriter.write(String.format("%.2f %.2f %d %.14f %.14f %.14f %.14f", lat, lon, year, lpjWinterWheatFrac, lpjSpringWheatFrac, lpjCornFrac, lpjRiceFrac));
				cropFractWriter.newLine();
			}
		}
		catch (IOException e) {
			LogWriter.print(e);
		}
		finally {
			if (landCoverWriter != null) {
				try {
					landCoverWriter.close();
				} 
				catch (IOException e) {
					LogWriter.print(e);
				}
			}
			if (cropFractWriter != null) {
				try {
					cropFractWriter.close();
				} 
				catch (IOException e) {
					LogWriter.print(e);
				}
			}
		}
	}
}
