package ac.ed.lurg.yield;

import java.io.File;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.FertiliserRate;
import ac.ed.lurg.types.IrrigationRate;
import ac.ed.lurg.types.YieldType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.WatchForFile;
import ac.sac.raster.AbstractTabularRasterReader;
import ac.sac.raster.RasterHeaderDetails;
import ac.sac.raster.RasterKey;

public class LPJYieldResponseMapReader {	

	private RasterHeaderDetails rasterProj;
	
	public LPJYieldResponseMapReader(RasterHeaderDetails rasterProj) {
		this.rasterProj = rasterProj; 
	}

	public YieldRaster getRasterData(Timestep timestep) {
		String rootDir = timestep.getYearSubDir(ModelConfig.YIELD_DIR);
		
		// wait for data to be available
		long startTime = System.currentTimeMillis();
		WatchForFile fileWatcher = new WatchForFile(new File(rootDir + File.separator + "done"));
		boolean foundFile = fileWatcher.await(ModelConfig.LPJG_MONITOR_TIMEOUT_SEC);
		if (!foundFile) {
			LogWriter.printlnError("Not able to find marker file.  May have timed out.");
			throw new RuntimeException("Not able to find marker file.  May have timed out.");
		}
		LogWriter.println("Found marker file in " + (System.currentTimeMillis() - startTime) + " ms");

		YieldRaster yieldRaster = new YieldRaster(rasterProj);
		readNppFile(yieldRaster, rootDir+File.separator+ModelConfig.ANPP_FILENAME);
		readYieldFile(yieldRaster, rootDir+File.separator+ModelConfig.YIELD_FILENAME);
		
		return yieldRaster;
	}
	
	private void readNppFile(YieldRaster yieldRaster, String fileToRead) {  // Passed in YieldRaster is updated
		
		AbstractTabularRasterReader<YieldResponsesItem> nppReader = new AbstractTabularRasterReader<YieldResponsesItem>("\\s+", 10, yieldRaster) {
			protected void setData(RasterKey key, YieldResponsesItem item, Map<String, Double> rowValues) {
				double adjFactor = 2 * 10 * ModelConfig.PASTURE_HARVEST_FRACTION;  // 2 for carbon units,  10 for kg/m2 to t/ha, and 0.5 for half grazed
								
				for (FertiliserRate fert : FertiliserRate.values()) {
					double pastureYield;
					if (ModelConfig.PASTURE_FERT_RESPONSE_FROM_LPJ)
						pastureYield = Math.max(getValueForCol(rowValues, LPJCropTypes.PASTURE_C3 + "N" + fert.getId()), getValueForCol(rowValues, LPJCropTypes.PASTURE_C4 + "N" + fert.getId())) * adjFactor;
					else
						pastureYield = Math.max(getValueForCol(rowValues, LPJCropTypes.PASTURE_C3.toString()), getValueForCol(rowValues, LPJCropTypes.PASTURE_C4.toString())) * adjFactor;

					item.setYield(YieldType.getYieldType(fert, IrrigationRate.NO_IRRIG), CropType.PASTURE, pastureYield);
					item.setYield(YieldType.getYieldType(fert, IrrigationRate.MAX_IRRIG), CropType.PASTURE, pastureYield);
				}
			}
		};
		
		nppReader.getRasterDataFromFile(fileToRead);
	}

	private void readYieldFile(YieldRaster yieldRaster, String fileToRead) {  	// Passed in YieldRaster is updated
		
		AbstractTabularRasterReader<YieldResponsesItem> yieldReader = new AbstractTabularRasterReader<YieldResponsesItem>("\\s+", 10, yieldRaster) {
			protected void setData(RasterKey key, YieldResponsesItem item, Map<String, Double> rowValues) {
				double adjFactor = 10;  // 10 for kg/m2 to t/ha
	
				String highFertString = YieldType.getYieldType(FertiliserRate.MAX_FERT, IrrigationRate.NO_IRRIG).getFertIrrigString();
	
				boolean isSpringWheat = (getValueForCol(rowValues, LPJCropTypes.SPRING_WHEAT + highFertString) > getValueForCol(rowValues, LPJCropTypes.WINTER_WHEAT + highFertString));
				item.setIsSpringWheat(isSpringWheat);
							
				for (FertiliserRate fert : FertiliserRate.values()) {
					for (IrrigationRate irrig : IrrigationRate.values()) {
						YieldType yieldType = YieldType.getYieldType(fert, irrig);
						String fertIrrigString = yieldType.getFertIrrigString();
						
						/*
						 * WARNING:  if you change mapping below you also need to change LpjgOutputer and IrrigationMaxAmountReader to match
						 */
						double ww = getValueForCol(rowValues, LPJCropTypes.WINTER_WHEAT + fertIrrigString) * adjFactor;
						double sw = getValueForCol(rowValues, LPJCropTypes.SPRING_WHEAT + fertIrrigString) * adjFactor;
						double co = getValueForCol(rowValues, LPJCropTypes.MAIZE + fertIrrigString) * adjFactor;
						double ri = getValueForCol(rowValues, LPJCropTypes.RICE + fertIrrigString) * adjFactor;
						
						setCalibratedYield(item, yieldType, CropType.WHEAT, isSpringWheat ? sw : ww);
						setCalibratedYield(item, yieldType, CropType.MAIZE, co);
						setCalibratedYield(item, yieldType, CropType.RICE, ri);
						setCalibratedYield(item, yieldType, CropType.OILCROPS, isSpringWheat ? sw : ww);
						setCalibratedYield(item, yieldType, CropType.PULSES, sw);
						setCalibratedYield(item, yieldType, CropType.STARCHY_ROOTS, sw);
					}
				}
			}
		};
		
		yieldReader.getRasterDataFromFile(fileToRead);
	}
	
	private void setCalibratedYield(YieldResponsesItem item, YieldType yieldType, CropType crop, double yield) {
		item.setYield(yieldType, crop, yield * crop.getCalibrationFactor());  // applying the calibration factor can't go in YieldResponsesItem.setYield as this is also called from elsewhere
	}
}